#extension GL_OES_standard_derivatives : enable

precision highp float;

#define PI 6.28318530718
#define TRANSPARENT vec4(0.)
#define WHITE vec4(1.)

#define NEWTON_ITERATIONS 4
#define NEWTON_MIN_SLOPE 0.001
#define SUBDIVISION_PRECISION 0.0000001
#define SUBDIVISION_MAX_ITERATIONS 10
#define kSplineTableSize 11
#define kSampleStepSize 1.0 / (11.0 - 1.0)

varying vec2 textureCoordinate;

uniform sampler2D inputImageTexture;
//uniform sampler2D inputImageTexture2;

uniform vec2 iResolution;
uniform float uParams[9];

//prequel_film_0相关
vec4 gaussianBlurFast(sampler2D tex, vec2 uv, vec2 resolution, float size) {
    // GAUSSIAN BLUR SETTINGS {{{
    float Directions = 16.0;// BLUR DIRECTIONS (Default 16.0 - More is better but slower)
    float Quality = 4.0;// BLUR QUALITY (Default 4.0 - More is better but slower)
    float Size = size;// BLUR SIZE (Radius)
    // GAUSSIAN BLUR SETTINGS }}}
    vec2 Radius = Size/resolution.xy;
    vec4 Color = texture2D(tex, uv);
    int count = 1;
    if (Size > 0.0) {
        for (float d=0.0; d< PI; d+= PI / Directions) {
            for (float i=1.0/Quality; i<=1.0; i+=1.0/Quality) {
                Color.rgb += texture2D(tex, uv+vec2(cos(d), sin(d))*Radius*i).rgb;
                count++;
            }
        }
    }
    Color.rgb /= float(count);
    return Color;
}
//prequel_film_1相关
vec3 shiftTexture(sampler2D tex, in vec2 size, in vec2 uv, const int x, const int y) {
    uv = uv * size.xy;
    uv = (uv + vec2(float(x), float(y))) / size.xy;
    return texture2D(tex, uv).rgb;
}

vec4 sharpenFilter(sampler2D tex, in vec2 size, in vec2 uv, float strength){
    vec3 f =
    shiftTexture(tex, size, uv, -1, -1) *  -1. +
    shiftTexture(tex, size, uv, 0, -1) *  -1. +
    shiftTexture(tex, size, uv, 1, -1) *  -1. +
    shiftTexture(tex, size, uv, -1, 0) *  -1. +
    shiftTexture(tex, size, uv, 0, 0) *   9. +
    shiftTexture(tex, size, uv, 1, 0) *  -1. +
    shiftTexture(tex, size, uv, -1, 1) *  -1. +
    shiftTexture(tex, size, uv, 0, 1) *  -1. +
    shiftTexture(tex, size, uv, 1, 1) *  -1.;
    vec4 src = texture2D(tex, uv);
    return vec4(mix(src.rgb, f, strength), src.a);
}

//prequel_film相关
void blend(in vec4 src, inout vec4 res, float opacity) {
    res = vec4(src.rgb*opacity*(1.-step(src.a, 0.))+ (1.-src.a*opacity)*res.rgb, max(res.a, src.a*opacity));
}
void blend(in vec4 src, inout vec4 res) {
    blend(src, res, 1.);
}

mat4 brightnessMatrix(float brightness){
    return mat4(1., 0., 0., 0., 0., 1., 0., 0., 0., 0., 1., 0., brightness, brightness, brightness, 1.);
}

float sampleValues[kSplineTableSize];
float A(float aA1, float aA2) {
    return 1.0 - 3.0 * aA2 + 3.0 * aA1;
}
float B(float aA1, float aA2) {
    return 3.0 * aA2 - 6.0 * aA1;
}
float C(float aA1) {
    return 3.0 * aA1;
}
float GetSlope(float aT, float aA1, float aA2) {
    return 3.0 * A(aA1, aA2)*aT*aT + 2.0 * B(aA1, aA2) * aT + C(aA1);
}
float CalcBezier(float aT, float aA1, float aA2) {
    return ((A(aA1, aA2)*aT + B(aA1, aA2))*aT + C(aA1))*aT;
}
float BinarySubdivide (float aX, float aA, float aB, float mX1, float mX2) {
    float currentX, currentT = 0.0;
    int i = 0;
    do {
        currentT = aA + (aB - aA) / 2.0;
        currentX = CalcBezier(currentT, mX1, mX2) - aX;
        if (currentX > 0.0) {
            aB = currentT;
        } else {
            aA = currentT;
        }
    } while (abs(currentX) > SUBDIVISION_PRECISION && ++i < SUBDIVISION_MAX_ITERATIONS);
    return currentT;
}
float NewtonRaphsonIterate (float aX, float aGuessT, float mX1, float mX2) {
    for (int i = 0; i < NEWTON_ITERATIONS; ++i) {
        float currentSlope = GetSlope(aGuessT, mX1, mX2);
        if (currentSlope == 0.0) {
            return aGuessT;
        }
        float currentX = CalcBezier(aGuessT, mX1, mX2) - aX;
        aGuessT -= currentX / currentSlope;
    }
    return aGuessT;
}
float GetTForX(float aX, float mX1, float mX2) {
    for (int i = 0; i < kSplineTableSize; ++i) {
        sampleValues[i] = CalcBezier(float(i) * kSampleStepSize, mX1, mX2);
    }
    float intervalStart = 0.0;
    int currentSample = 1;
    int lastSample = kSplineTableSize - 1;
    for (; currentSample != lastSample && sampleValues[currentSample] <= aX; ++currentSample) {
        intervalStart += kSampleStepSize;
    }
    --currentSample;
    // Interpolate to provide an initial guess for t
    float dist = (aX - sampleValues[currentSample]) / (sampleValues[currentSample + 1] - sampleValues[currentSample]);
    float guessForT = intervalStart + dist * kSampleStepSize;
    float initialSlope = GetSlope(guessForT, mX1, mX2);
    if (initialSlope >= NEWTON_MIN_SLOPE) {
        return NewtonRaphsonIterate(aX, guessForT, mX1, mX2);
    } else if (initialSlope == 0.0) {
        return guessForT;
    } else {
        return BinarySubdivide(aX, intervalStart, intervalStart + kSampleStepSize, mX1, mX2);
    }
}
/*
 * @param aX: 传入时间变量
 * @param mX1/mY1/mX2/mY2: 贝塞尔曲线四个值
 * 说明: 这个函数以上的其他函数都是本函数使用的辅助函数
 */
float KeySpline(float aX, float mX1, float mY1, float mX2, float mY2) {
    if (mX1 == mY1 && mX2 == mY2) return aX;// linear
    return CalcBezier(GetTForX(aX, mX1, mX2), mY1, mY2);
}
vec3 rgb2hsl(in vec3 c) {
    float H, S, L;
    float var_R = c.r;
    float var_G = c.g;
    float var_B = c.b;
    float var_Min = min(min(var_R, var_G), var_B);//Min. value of RGB
    float var_Max = max(max(var_R, var_G), var_B);//Max. value of RGB
    float del_Max = var_Max - var_Min;//Delta RGB value
    L = (var_Max + var_Min)/ 2.;
    if (del_Max == 0.)  { //This is a gray, no chroma...
        H = 0.;
        S = 0.;
    } else { //Chromatic data...
        if (L < 0.5) S = del_Max / (var_Max + var_Min);
        else S = del_Max / (2. - var_Max - var_Min);

        float del_R = (((var_Max - var_R) / 6.) + (del_Max / 2.)) / del_Max;
        float del_G = (((var_Max - var_G) / 6.) + (del_Max / 2.)) / del_Max;
        float del_B = (((var_Max - var_B) / 6.) + (del_Max / 2.)) / del_Max;

        if (var_R == var_Max) H = del_B - del_G;
        else if (var_G == var_Max) H = (1. / 3.) + del_R - del_B;
        else if (var_B == var_Max) H = (2. / 3.) + del_G - del_R;

        if (H < 0.) H += 1.;
        if (H > 1.) H -= 1.;
    }
    return vec3(H, S, L);
}
float hue2rgb(float v1, float v2, float vH){
    if (vH < 0.) vH += 1.;
    if (vH > 1.) vH -= 1.;
    if ((6. * vH) < 1.) return (v1 + (v2 - v1) * 6. * vH);
    if ((2. * vH) < 1.) return (v2);
    if ((3. * vH) < 2.) return (v1 + (v2 - v1) * ((2. / 3.) - vH) * 6.);
    return v1;
}
vec3 hsl2rgb(in vec3 c) {
    float R, G, B;
    float H = c.x;
    float S = c.y;
    float L = c.z;
    if (S == 0.) {
        return vec3(L);
    } else {
        float var_2;
        if (L < 0.5) var_2 = L * (1. + S);
        else var_2 = (L + S) - (S * L);
        float var_1 = 2. * L - var_2;
        R = hue2rgb(var_1, var_2, H + (1. / 3.));
        G = hue2rgb(var_1, var_2, H);
        B = hue2rgb(var_1, var_2, H - (1. / 3.));
    }
    return vec3(R, G, B);
}

void main() {
    vec2 uv = textureCoordinate;

    vec4 source = texture2D(inputImageTexture, uv);

    //prequel_film_0
    vec4 film0 = gaussianBlurFast(inputImageTexture, uv, iResolution, 10.*uParams[1]);
    //prequel_film_1
    vec4 film1 = sharpenFilter(inputImageTexture, iResolution, uv, uParams[2]);
    vec4 film12Concat = mix(film0, film1, 0.2);

    //prequel_film_2
    vec4 film2 = film12Concat;//texture2D(inputImageTexture2, uv);
    film2.rgb = clamp(film2.rgb-22./255., 0., 1.);
    film2.r = clamp(242./247.*(film2.r-8./255.), 0., 1.);
    film2.g = clamp(250./245.*(film2.g-10./255.), 0., 1.);
    film2.b = clamp(239./238.*(film2.b+16./255.), 0., 1.);

    //prequel_film
    vec4 color = WHITE;
    vec4 layer = source;//texture2D(inputImageTexture, uv);
    vec4 filterLayer = film2;//texture2D(inputImageTexture2, uv);
    vec4 rgbshift = texture2D(inputImageTexture, uv+vec2(.004, .004)*(uParams[4]*2. - 1.));
    filterLayer.r = mix(filterLayer.r, rgbshift.r, uParams[5]);
    filterLayer.g = mix(filterLayer.g, rgbshift.g, uParams[6]);
    filterLayer.b = mix(filterLayer.b, rgbshift.b, uParams[7]);
    filterLayer = brightnessMatrix(.6*uParams[3]-.3)*filterLayer;

    //阴影
    vec3 hsl = rgb2hsl(filterLayer.rgb);
    float pct = hsl.z;
    float pct2 = uParams[8];
    float newL;
    if (pct < 0. || pct > .5) newL = pct;
    else newL = KeySpline(pct*2., pct2*.4, .4-pct2*.4, 1., 1.)*.5;
    filterLayer.rgb = hsl2rgb(vec3(hsl.x, hsl.y, newL));

    blend(filterLayer, layer, uParams[0]);
    blend(layer, color);
    gl_FragColor = color;
}